# This file is part of CloudControl.
#
# CloudControl is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CloudControl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CloudControl.  If not, see <http://www.gnu.org/licenses/>.


""" Manage the persistence of jobs on disk.
"""

import os
import shutil
import glob
import cPickle as pickle


class JobsStore(object):

    def __init__(self, base_directory):
        self._base_directory = base_directory

        # Create the base directory if it doesn't exists yet:
        try:
            os.mkdir(base_directory)
        except OSError as err:
            if err.errno != 17:
                raise

    def _get_job_dir(self, job_id):
        return os.path.join(self._base_directory, job_id)

    def _get_attrs_dir(self, job_id):
        return os.path.join(self._get_job_dir(job_id), 'job.dat')

    def _get_counter_file(self):
        return os.path.join(self._base_directory, 'counter')

    def _get_attachment_file(self, job_id, name):
        return os.path.join(self._get_job_dir(job_id), '%s.attachment' % name)

    def store_id_counter(self, current):
        with open(self._get_counter_file(), 'w') as fcounter:
            pickle.dump(current, fcounter)

    def get_id_counter(self):
        try:
            with open(self._get_counter_file(), 'r') as fcounter:
                return pickle.load(fcounter)
        except IOError as err:
            if err.errno != 2:
                raise
        return 0

    def create_job(self, job):
        """ Create a new job.
        """
        try:
            os.mkdir(self._get_job_dir(job.id))
        except OSError as err:
            if err.errno != 17:
                raise
        else:
            self.update_job(job)

    def update_job(self, job):
        """ Update attributes of a job.
        """
        with open(self._get_attrs_dir(job.id), 'w') as fjob:
            pickle.dump(job, fjob)

    def delete_job(self, job_id):
        """ Delete a job.
        """
        shutil.rmtree(self._get_job_dir(job_id))

    def iter_jobs(self):
        """ Iterate over stored jobs.
        """
        for filename in glob.glob(os.path.join(self._base_directory, 'job-*', 'job.dat')):
            with open(filename, 'r') as fjob:
                yield pickle.load(fjob)

    def get_attachment_file(self, job_id, name, append=False):
        """ Get an open-writable file object for an attachment.
        """
        mode = 'a' if append else 'w'
        return open(self._get_attachment_file(job_id, name), mode)

    def read_attachment(self, job_id, name):
        """ Get the content of an attachment in a string.
        """
        with open(self._get_attachment_file(job_id, name), 'r') as fattach:
            return fattach.read()  #FIXME: DANGEROUS IF THE FILE IS BIG LIKE UR MOM
